"""
Initial Mass Functions

Contains some predefined IMFs from different papers/authors:

* Salpeter 1955
* Miller & Scalo 1979
* Ferrini, Palla & Penco 1998
* Starburst 1999
* Kroupa 2002
* Chabrier 2003
* Maschberger 2012

and a way to define new functions subclassing IMF

"""
import math
import scipy.integrate
import starmatrix.settings


def select_imf(name, params={}):
    imfs = {
        "salpeter": Salpeter,
        "starburst": Starburst,
        "chabrier": Chabrier,
        "ferrini": Ferrini,
        "kroupa": Kroupa,
        "miller_scalo": MillerScalo,
        "maschberger": Maschberger
    }
    return imfs[name](params)


class IMF:
    def __init__(self, params={}):
        self.params = params
        self.set_mass_limits()
        self.normalization_factor = 1.0 / self.integrated_m_phi_in_mass_interval()
        self.stars_per_mass_unit = self.normalization_factor * self.integrated_phi_in_mass_interval()
        self.set_params()

    def integrated_m_phi_in_mass_interval(self):
        return scipy.integrate.quad(self.m_phi, self.m_low, self.m_up)[0]

    def integrated_phi_in_mass_interval(self):
        return scipy.integrate.quad(self.phi, self.m_low, self.m_up)[0]

    def for_mass(self, m):
        """
        The value of (m * imf) normalized so integral(m * imf) = 1 in [m_low, m_up]

        """
        if m <= 0:
            return 0.0

        return self.normalization_factor * self.m_phi(m)

    def set_mass_limits(self):
        if "imf_m_low" in self.params:
            self.m_low = self.params["imf_m_low"]
        else:
            self.m_low = starmatrix.settings.default["imf_m_low"]

        if "imf_m_up" in self.params:
            self.m_up = self.params["imf_m_up"]
        else:
            self.m_up = starmatrix.settings.default["imf_m_up"]

    def set_params(self):
        pass

    def m_phi(self, m):
        return m

    def phi(self, m):
        return self.m_phi(m) / m

    def description(self):
        return "Base Initial Mass Function class"


class Salpeter(IMF):
    def m_phi(self, m):
        return m * (m ** -(self.alpha()))

    def alpha(self):
        if "imf_alpha" in self.params:
            return self.params["imf_alpha"]
        else:
            return starmatrix.settings.default["imf_alpha"]

    def description(self):
        return "IMF from Salpeter 1955"


class Starburst(Salpeter):

    def set_mass_limits(self):
        self.m_low = 1.0
        self.m_up = 120.0

    def description(self):
        return "IMF from Starburst 1999"


class MillerScalo(IMF):
    def m_phi(self, m):
        return math.exp(-((math.log10(m) + 1.02) ** 2) / (2 * (0.68 ** 2)))

    def description(self):
        return "IMF from Miller & Scalo 1979"


class Ferrini(IMF):
    def m_phi(self, m):
        return 10 ** (-math.sqrt(0.73 + math.log10(m) * (1.92 + math.log10(m) * 2.07))) / m ** 0.52

    def description(self):
        return "IMF Ferrini, Palla & Penco 1998"


class Kroupa(IMF):
    def m_phi(self, m):
        if 0.015 <= m < 0.08:
            return m * (m ** -0.35)
        elif 0.08 <= m < 0.5:
            return m * 0.08 * (m ** -1.3)
        elif 0.5 <= m < 1.0:
            return m * 0.04 * (m ** -2.3)
        elif 1 <= m:
            return m * 0.04 * (m ** -2.7)
        else:
            return 0

    def description(self):
        return "IMF from Kroupa 2002"


class Chabrier(IMF):
    def m_phi(self, m):
        if m <= 1:
            return 0.086*math.exp(-((math.log10(m) - math.log10(0.22))**2)/(2*(0.57**2)))
        else:
            return m*0.043*(m**-2.35)

    def description(self):
        return "IMF from Chabrier 2003"


class Maschberger(IMF):
    def set_mass_limits(self):
        self.m_low = 0.15
        self.m_up = 100.0

    def m_phi(self, m):
        return m * self.a() * \
               (self.m_mu(m) ** (-self.aalfa())) * \
               ((1 + (self.m_mu(m) ** (1 - self.aalfa()))) ** (-self.beta()))

    def m_mu(self, m):
        return m / self.mu()

    def g1(self):
        return (1 + ((0.15 / 0.2) ** (1 - self.aalfa()))) ** (1 - self.beta())

    def g2(self):
        return (1 + ((100 / 0.2) ** (1 - self.aalfa()))) ** (1 - self.beta())

    def a(self):
        return ((1 - self.aalfa()) * (1 - self.beta()) / self.mu()) * (1 / (self.g2() - self.g1()))

    def mu(self):
        return 0.2

    def aalfa(self):
        return 2.3

    def beta(self):
        return 1.4

    def description(self):
        return "IMF from Maschberger 2012"



import pytest
import numpy as np
import starmatrix.settings as settings
from starmatrix.dtds import select_dtd
from starmatrix.dtds import dtd_correction
from starmatrix.dtds import dtd_ruiz_lapuente
from starmatrix.dtds import dtd_maoz_graur
from starmatrix.dtds import dtd_castrillo
from starmatrix.dtds import dtd_greggio
from starmatrix.dtds import dtd_close_dd_04
from starmatrix.dtds import dtd_close_dd_1
from starmatrix.dtds import dtd_wide_dd_04
from starmatrix.dtds import dtd_wide_dd_1
from starmatrix.dtds import dtd_sd_chandra
from starmatrix.dtds import dtd_sd_subchandra
from starmatrix.dtds import dtd_chen
from starmatrix.dtds import dtds_strolger


@pytest.fixture
def available_dtds():
    """
    Fixture returning the names of all available DTDs defined in settings
    """
    return settings.valid_values["dtd_sn"]


def test_dtds_presence(available_dtds):
    for dtd in available_dtds:
        assert select_dtd(dtd) is not None


def test_select_dtd(available_dtds):
    dtds = [dtd_ruiz_lapuente, dtd_maoz_graur, dtd_castrillo, dtd_greggio, dtd_chen,
            dtd_close_dd_04, dtd_close_dd_1, dtd_wide_dd_04, dtd_wide_dd_1, dtd_sd_chandra, dtd_sd_subchandra,
            dtds_strolger["fit_1"], dtds_strolger["fit_2"], dtds_strolger["fit_3"], dtds_strolger["fit_4"], dtds_strolger["optimized"]]

    assert len(available_dtds) == len(dtds)

    for i in range(len(available_dtds)):
        times = [0, 0.001, 0.04, 0.1, 0.4, 1, 2, 9.] + list(np.random.rand(5)) + list(np.random.rand(5) * 9)
        for time in times:
            assert select_dtd(available_dtds[i])(time) == dtds[i](time)


def test_no_negative_time_values():
    t = -1
    assert dtd_ruiz_lapuente(t) == 0.0
    assert dtd_maoz_graur(t) == 0.0
    assert dtd_close_dd_04(t) == 0.0
    assert dtd_close_dd_1(t) == 0.0
    assert dtd_wide_dd_04(t) == 0.0
    assert dtd_wide_dd_1(t) == 0.0
    assert dtd_sd_chandra(t) == 0.0
    assert dtd_sd_subchandra(t) == 0.0


def test_dtd_correction_factor():
    assert dtd_correction({}) == 1.0
    assert dtd_correction({'dtd_correction_factor': 3.0}) == 3.0



import math

import starmatrix.constants as constants


def value_in_interval(value, interval=[]):
    return min(max(interval[0], value), interval[1])


def secondary_mass_fraction(mu):
    """
    Distribution function of the mass fraction of the secondary in binary systems / SNI
    mu = Mass_secondary / Mass_binary_system
    From: Matteucci, F. & Greggio, L. 1986, A&A, 154, 279
    with Gamma = 2 as Greggio, L., Renzini, A.: 1983a, Astron. Astrophys. 118, 217

    """
    gamma = 2.0
    return (2.0 ** (1.0 + gamma)) * (1.0 + gamma) * (mu ** gamma)


def tau_polinomyal_coefficients(z):
    """
    Coefficients (z-dependent) for the log(tau) formula from
    Raiteri C.M., Villata M. & Navarro J.F., 1996, A&A 315, 105-115

    """
    log_z = math.log10(z)
    log_z_2 = log_z ** 2

    a0 = 10.13 + 0.07547 * log_z - 0.008084 * log_z_2
    a1 = -4.424 - 0.7939 * log_z - 0.1187 * log_z_2
    a2 = 1.262 + 0.3385 * log_z + 0.05417 * log_z_2

    return [a0, a1, a2]


def stellar_lifetime(stellar_m, z):
    """
    Empirical formula for stellar lifetimes from
    Raiteri C.M., Villata M. & Navarro J.F., 1996, A&A 315, 105-115

    """
    log_m = math.log10(stellar_m)
    a0, a1, a2 = tau_polinomyal_coefficients(z)

    log_tau = a0 + a1 * log_m + a2 * (log_m ** 2)

    return math.pow(10, log_tau - 9)


def stellar_mass(tau, z):
    """
    Derived from the stellar lifetimes formula from
    Raiteri C.M., Villata M. & Navarro J.F., 1996, A&A 315, 105-115
    solving the equation for the log(M).
    This function returns always the smaller root, as that is the
    good fit for masses up to the max_mass_allowed(z)

    """
    log_tau = math.log10(tau * 1e9)  # years to Gyrs
    a0, a1, a2 = tau_polinomyal_coefficients(z)
    square = math.sqrt((a1 ** 2) - (4 * a2 * (a0 - log_tau)))
    log_mass_minus = (-a1 - square) / (2 * a2)

    return round(math.pow(10, log_mass_minus), 10)


def max_mass_allowed(z):
    """
    The formula for stellar lifetimes from Raiteri et al is a good fit up until
    a (dependent on z) critical mass. After it tau increases and we consider it non valid.

    """
    log_z = math.log10(z)
    log_z_2 = log_z ** 2
    _, a1, a2 = tau_polinomyal_coefficients(z)
    return float(math.floor((math.pow(10, -a1/(2 * a2)))))


def total_energy_ejected(t):
    """
    Thermal and kinetic energy released by each type of SN up to the time t after the explosion
    where tc is the characteristic cooling time of the shell sorrounding the remnant (53000 yrs)
    from Ferrini & Poggiantti, 1993, ApJ, 410, 44F

    """
    if t <= 0:
        return 0.0

    tc = 5.3e-5
    if t > tc:
        rt = (tc / t) ** 0.4
        return 1 - 0.44 * (rt ** 2) * (1 - 0.41 * rt) - 0.22 * (rt ** 2)
    else:
        return 9811.32 * t


def newton_cotes(a, b, f):
    """
    Integration using Newton-Cotes formula with degree 6 (7 points)

    """
    NEWTON_COTES_POINTS = 7
    NEWTON_COTES_COEFFICIENTS = [0.29285714, 1.54285714, 0.19285714, 1.94285714, 0.19285714, 1.54285714, 0.29285714]

    h = (b - a) / (NEWTON_COTES_POINTS - 1)
    sum_fs = 0.0
    for i in range(0, NEWTON_COTES_POINTS):
        sum_fs += NEWTON_COTES_COEFFICIENTS[i] * f(a + (i * h))

    return h * sum_fs


def imf_binary_primary(m, imf, binary_fraction=constants.BIN_FRACTION):
    """
    Initial mass function for primary stars of binary systems
    Integrated between  m' and m'' using Newton-Cotes
    Returns 0 unless m is in (1.5, 16)

    """
    m_inf = max(constants.B_MIN, m)
    m_sup = min(constants.B_MAX, 2 * m)
    if m <= 0 or m_sup <= m_inf:
        return 0.0

    return binary_fraction * newton_cotes(m_inf, m_sup, phi_primary(m, imf))


def imf_binary_secondary(m, imf, SNI_events=False, binary_fraction=constants.BIN_FRACTION):
    """
    Initial mass function for secondary stars of binary systems
    Optionally ocurring Supernova I events
    Integrated between  m' and m'' using Newton-Cotes
    If SNI_events = False then returns 0 unless m is in (0, 8)

    """
    m_inf = max(constants.B_MIN, 2 * m)
    m_sup = constants.B_MAX
    if SNI_events:
        m_sup = min(constants.B_MAX, constants.M_SNII + m)

    if m <= 0 or m_sup <= m_inf:
        return 0.0

    return binary_fraction * newton_cotes(m_inf, m_sup, phi_secondary(m, imf))


def imf_zero(m, imf, binary_fraction=constants.BIN_FRACTION):
    """
    Initial mass function for stars that are single or
    part of binary systems not giving rise to SN I events

    """
    if constants.B_MIN <= m <= constants.B_MAX:
        return imf.for_mass(m) * (1.0 - binary_fraction)
    else:
        return imf.for_mass(m)


def global_imf(m, imf, binary_fraction=constants.BIN_FRACTION):
    """
    global initial mass function from Ferrini et al.*,1992, ApJ, 387, 138

    """
    if m < constants.M_MIN:
        return 0.0
    if constants.M_MIN <= m < 1.5:
        return imf_zero(m, imf, binary_fraction) + imf_binary_secondary(m, imf, binary_fraction)
    elif 1.5 <= m < 8:
        return imf_zero(m, imf, binary_fraction) + imf_binary_secondary(m, imf, binary_fraction) + imf_binary_primary(m, imf, binary_fraction)
    elif 8 <= m < 16:
        return imf_zero(m, imf, binary_fraction) + imf_binary_primary(m, imf, binary_fraction)
    elif 16 <= m:
        return imf_zero(m, imf, binary_fraction)


def phi_primary(m, imf):
    """
    Expression to integrate for each mass m for the IMF for primary stars of binary systems

    """
    return lambda binary_mass: secondary_mass_fraction(1.0 - (m / binary_mass)) * \
        imf.for_mass(binary_mass) * \
        m / (binary_mass ** 2)


def phi_secondary(m, imf):
    """
    Expression to integrate for each mass m for the IMF for secondary stars of binary systems

    """
    return lambda binary_mass: secondary_mass_fraction(m / binary_mass) * \
        imf.for_mass(binary_mass) * \
        m / (binary_mass ** 2)


def imf_supernovae_II(m, imf, binary_fraction=constants.BIN_FRACTION):
    if m > constants.M_SNII:
        return (imf_zero(m, imf, binary_fraction) + imf_binary_primary(m, imf, binary_fraction)) / m
    else:
        return 0


def return_fraction(m_inf, m_sup, expelled, imf, binary_fraction=constants.BIN_FRACTION):
    r = newton_cotes(
        m_inf,
        m_sup,
        lambda m:
            (global_imf(m, imf, binary_fraction) / m) * (m - expelled.for_mass(m)['remnants'])
        )

    return r



import pytest
from pytest_mock import mocker
import math
import numpy as np
from starmatrix.elements import Expelled
import starmatrix.settings as settings


@pytest.fixture
def expelled():
    """
    Fixture returning a Expelled instance initialized with the default expelled data file
    """
    expelled = Expelled(settings.default["expelled_elements_filename"])
    return expelled


def test_read_file_on_init(expelled):
    assert len(expelled.mass_points) == 80
    assert expelled.mass_points[0] == 0.8
    assert expelled.mass_points[-1] == 100
    assert len(expelled.by_mass[0.8]) == len(expelled.elements_list)


def test_interpolation_for_mass_data(expelled):
    i = np.random.randint(0, len(expelled.mass_points)-1)

    m_low = expelled.mass_points[i]
    m_up = expelled.mass_points[i + 1]
    m = m_low + 0.5 * (m_up - m_low)

    expelled_for_mass = expelled.for_mass(m)

    for element in expelled.elements_list:
        interval = sorted([expelled.by_mass[m_up][element] / m, expelled.by_mass[m_low][element] / m])

        assert expelled_for_mass[element] <= interval[1]
        assert expelled_for_mass[element] >= interval[0]


def test_extrapolation_for_mass_data(expelled):
    m_low = expelled.mass_points[-2]
    m_up = expelled.mass_points[-1]
    last_step_data_size = m_up - m_low

    m = m_up + last_step_data_size

    expelled_for_mass = expelled.for_mass(m)

    for element in expelled.elements_list:
        by_mass_m_up = expelled.by_mass[m_up][element]
        by_mass_m_low = expelled.by_mass[m_low][element]
        extrapolation = by_mass_m_up + (by_mass_m_up - by_mass_m_low)

        assert expelled_for_mass[element] == extrapolation / m


def test_for_mass_with_no_yield_corrections(expelled):
    elements = expelled.elements_list
    no_correction = dict(zip(elements, np.ones(len(elements))))

    m = np.random.rand() * 40

    assert expelled.for_mass(m) == expelled.for_mass(m, {})
    assert expelled.for_mass(m) == expelled.for_mass(m, no_correction)


def test_for_mass_with_yield_corrections(expelled):
    elements = expelled.elements_list
    m = np.random.rand() * 40

    zero_correction = dict(zip(elements, np.zeros(len(elements))))
    zero_corrected = expelled.for_mass(m, zero_correction)
    assert zero_corrected['mass'] == m
    for e in elements:
        assert zero_corrected[e] == 0

    no_correction = expelled.for_mass(m)
    for element_corrected in elements:
        element_correction = {element_corrected: 2.57}
        corrected = expelled.for_mass(m, element_correction)
        for element in elements:
            if element == element_corrected:
                assert corrected[element] == no_correction[element] * 2.57
            else:
                assert corrected[element] == no_correction[element]


def test_cri_lim_exception(mocker):
    expelled = Expelled(settings.default["expelled_elements_filename"])
    assert expelled.cri_lim_yields is False

    mocker.patch.object(Expelled, "read_expelled_elements_file")
    cri_lim_expelled = Expelled("expelled_CRI-LIM-elements_filename")
    assert cri_lim_expelled.cri_lim_yields is True



import pytest
import numpy as np
import starmatrix.functions as functions
import starmatrix.constants as constants
import starmatrix.settings as settings
import starmatrix.elements as elements
from starmatrix.imfs import select_imf


def test_value_in_interval():
    interval_min = 1.0
    interval_max = 100
    interval = [interval_min, interval_max]
    value_in = 25
    value_out_min = 0.8
    value_out_max = 101

    assert functions.value_in_interval(value_in, interval) == value_in
    assert functions.value_in_interval(value_out_min, interval) == interval_min
    assert functions.value_in_interval(value_out_max, interval) == interval_max


def test_secondary_mass_fraction():
    for m in [0.33, 3.33, 33.7, 73.0]:
        assert functions.secondary_mass_fraction(m) == 24 * m ** 2


def test_mean_lifetime_stellar_mass_relation():
    z = 0.02
    stellar_mass_test = 4.0
    lifetime_test = 0.15

    stellar_mass = functions.stellar_mass(lifetime_test, z)
    lifetime = functions.stellar_lifetime(stellar_mass_test, z)

    assert np.isclose(functions.stellar_mass(lifetime, z), stellar_mass_test,  rtol=0.005)
    assert np.isclose(functions.stellar_lifetime(stellar_mass, z), lifetime_test, rtol=0.005)


def test_total_energy_no_negative_time_values():
    t = -1
    assert functions.total_energy_ejected(t) == 0.0

    assert functions.total_energy_ejected(1e-5) > 0.0

    t = functions.stellar_lifetime(5, 0.02)
    assert functions.total_energy_ejected(t) > 0.0


def test_imf_zero():
    m_in_binaries_range = 5.0
    m_lower = constants.B_MIN - 0.5
    m_up = constants.B_MAX + 0.5
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)

    assert functions.imf_zero(m_lower, imf) == imf.for_mass(m_lower)
    assert functions.imf_zero(m_up, imf) == imf.for_mass(m_up)

    imf_bin = imf.for_mass(m_in_binaries_range) * (1.0 - constants.BIN_FRACTION)
    assert functions.imf_zero(m_in_binaries_range, imf) == imf_bin


def test_imf_binaries_are_zero_for_non_valid_masses():
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)

    assert functions.imf_binary_primary(-1, imf) == 0.0
    assert functions.imf_binary_primary(constants.M_MIN / 2, imf) == 0.0
    assert functions.imf_binary_secondary(-1, imf) == 0.0
    assert functions.imf_binary_secondary(constants.B_MAX * 2, imf) == 0.0


def test_imf_binary_primary_integrates_phi_primary():
    m_in_binaries_range = 5.0
    m_sup = 2 * m_in_binaries_range
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)

    expected = settings.default['binary_fraction'] * \
        functions.newton_cotes(m_in_binaries_range, m_sup, functions.phi_primary(m_in_binaries_range, imf))
    assert functions.imf_binary_primary(m_in_binaries_range, imf) == expected


def test_imf_binary_secondary_integrates_phi_secondary():
    m_in_binaries_range = 5.0
    m_inf = 2 * m_in_binaries_range
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)

    expected = settings.default['binary_fraction'] * \
        functions.newton_cotes(m_inf, constants.B_MAX, functions.phi_secondary(m_in_binaries_range, imf))
    assert functions.imf_binary_secondary(m_in_binaries_range, imf) == expected


def test_global_imf():
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)

    assert functions.global_imf(constants.M_MIN - 0.001, imf) == 0
    assert functions.global_imf(100, imf) == functions.imf_zero(100, imf)
    for m in [1, 4, 8, 10, 40]:
        assert 0 < functions.global_imf(m, imf)
        assert functions.imf_zero(100, imf) < functions.global_imf(m, imf)


def test_imf_supernovae_II_non_zero_for_SNII_masses():
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)

    assert functions.imf_supernovae_II(constants.M_SNII + 0.01, imf) > 0
    assert functions.imf_supernovae_II(constants.M_SNII + np.random.sample() * 100, imf) > 0
    assert functions.imf_supernovae_II(constants.M_SNII + 100, imf) > 0


def test_imf_supernovae_II_is_zero_for_lower_masses():
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)

    assert functions.imf_supernovae_II(np.random.sample() * constants.M_SNII, imf) == 0
    assert functions.imf_supernovae_II(constants.M_SNII, imf) == 0


def test_imf_supernovae_II_includes_binary_primaries():
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)
    m = (constants.B_MIN + constants.B_MAX) / 2

    assert m > constants.M_SNII
    assert functions.imf_supernovae_II(m, imf) > functions.imf_zero(m, imf)/m


def test_return_fractions():
    imf = select_imf(np.random.choice(settings.valid_values["imf"]), settings.default)
    default_yields_file = settings.default['expelled_elements_filename']
    stellar_yields = elements.Expelled(default_yields_file)

    expected = functions.newton_cotes(
        1,
        100,
        lambda m:
            (functions.global_imf(m, imf)/m) * (m - stellar_yields.for_mass(m)['remnants'])
        )

    assert functions.return_fraction(1, 100, stellar_yields, imf) == expected



import argparse
import os
import shutil
import pytest
from pytest_mock import mocker
import starmatrix.cli as cli
import starmatrix.model as model
import starmatrix.settings as settings


@pytest.fixture
def deactivate_os_actions(mocker):
    """
    Fixture to disable all calls to the operating system (mkdir, rm, copy...)
    """
    mocker.patch.object(model, 'Model')
    mocker.patch.object(model.Model, 'run')
    mocker.patch.object(os, 'makedirs')
    mocker.patch.object(shutil, 'rmtree')
    mocker.patch.object(shutil, 'copy')
    mocker.patch.object(argparse.ArgumentParser, 'parse_args')
    argparse.ArgumentParser.parse_args.return_value = argparse.Namespace(generate_config=False, config=None)
    os.makedirs.return_value = True
    shutil.rmtree.return_value = True


@pytest.fixture
def mock_config_file(mocker):
    """
    Fixture mocking contents of a config file reading
    """
    config_file_content = "m_max: 33.0\ntotal_time_steps: 123"
    mocked_file = mocker.mock_open(read_data=config_file_content)
    mocker.patch.object(cli, 'open', mocked_file)
    mocker.spy(cli, "read_config_file")
    return {'m_max': 33.0, 'total_time_steps': 123}


@pytest.fixture
def output_dir_existence(mocker):
    mocker.patch.object(cli, 'exists')
    cli.exists.return_value = True


def test_option_generate_config(mocker, deactivate_os_actions):
    argparse.ArgumentParser.parse_args.return_value = argparse.Namespace(generate_config=True)
    mocker.spy(cli, "create_template_config_file")
    cli.main()
    cli.create_template_config_file.assert_called()


def test_option_config(mocker, deactivate_os_actions, mock_config_file):
    argparse.ArgumentParser.parse_args.return_value = argparse.Namespace(generate_config=False, config='ejectas.dat')
    cli.main()
    cli.read_config_file.assert_called_once_with('ejectas.dat')


def test_model_is_configured_properly(mocker, deactivate_os_actions, mock_config_file):
    argparse.ArgumentParser.parse_args.return_value = argparse.Namespace(generate_config=False, config='ejectas.dat')
    cli.main()
    expected_context = settings.validate(mock_config_file)

    model.Model.assert_called_once_with(expected_context)
    model.Model(expected_context).run.assert_called()


def test_model_is_run(mocker, deactivate_os_actions):
    cli.main()
    model.Model.assert_called()
    model.Model(settings.default_settings()).run.assert_called()


def test_creation_of_output_directory(mocker, deactivate_os_actions, mock_config_file):
    mocker.spy(cli, "create_output_directory")
    cli.main()
    cli.create_output_directory.assert_called_once_with('results')
    os.makedirs.assert_called_once_with('results')
    model.Model.assert_called()


def test_existent_output_directory(mocker, deactivate_os_actions, mock_config_file, output_dir_existence):
    cli.main()
    os.makedirs.assert_not_called()
    model.Model.assert_called()



import pytest
import random
import numpy as np
import starmatrix.supernovae as sn


def test_empty_yields_set_is_all_zeroes():
    empty_set = sn.empty_yields_set()

    assert len(empty_set) > 0
    assert len(empty_set) == len(sn.sn_elements_list)
    for element in sn.sn_elements_list:
        assert element in empty_set.keys()
        assert empty_set[element] == 0


def test_yields_structure():
    options = ["iwa1998", "sei2013", "ln2020", "ln2018-1", "ln2018-2", "ln2018-3",
               "br2019-1", "br2019-2", "gro2021-1", "gro2021-2", "mor2018-1", "mor2018-2"]
    for dataset in options:
        yields = sn.yields(dataset, 0)

        assert len(yields) > 0
        assert len(yields) == len(sn.sn_elements_list)
        for element in sn.sn_elements_list:
            assert element in yields.keys()


def test_yields_from_iwamoto():
    minus1 = sn.yields_from_iwamoto(-1)
    minus0301 = sn.yields_from_iwamoto(-0.301)
    minus001 = sn.yields_from_iwamoto(-0.01)
    plus05 = sn.yields_from_iwamoto(0.5)

    assert minus1 == minus0301
    assert minus0301 != minus001
    assert minus001 == plus05


def test_yields_from_seitenzahl():
    minus2 = sn.yields_from_seitenzahl(-2)
    minus15 = sn.yields_from_seitenzahl(-1.5)
    minus1 = sn.yields_from_seitenzahl(-1)
    minus065 = sn.yields_from_seitenzahl(-0.65)
    minus03 = sn.yields_from_seitenzahl(-0.3)
    minus015 = sn.yields_from_seitenzahl(-0.15)
    minus01 = sn.yields_from_seitenzahl(-0.1)
    plus1 = sn.yields_from_seitenzahl(1)

    assert minus2 == minus15
    assert minus15 != minus1
    assert minus1 == minus065
    assert minus065 != minus03
    assert minus03 == minus015
    assert minus015 != minus01
    assert minus01 == plus1


def test_yields_from_leung_nomoto_2018():
    leung_nomoto_sources = [
        sn.yields_from_leung_nomoto_2018_table6,
        sn.yields_from_leung_nomoto_2018_table8,
        sn.yields_from_leung_nomoto_2018_table10
    ]

    for ln_dataset in leung_nomoto_sources:
        minus2 = ln_dataset(-2)
        minus165 = ln_dataset(-1.65)
        minus1 = ln_dataset(-1)
        minus065 = ln_dataset(-0.65)
        minus03 = ln_dataset(-0.3)
        minus015 = ln_dataset(-0.15)
        zero = ln_dataset(0.0)
        plus015 = ln_dataset(0.15)
        plus03 = ln_dataset(0.3)
        plus039 = ln_dataset(0.39)
        plus047 = ln_dataset(0.47)
        plus059 = ln_dataset(0.59)
        plus069 = ln_dataset(0.69)
        plus1 = ln_dataset(1)

        assert minus2 == minus165
        assert minus165 != minus1
        assert minus1 == minus065
        assert minus065 != minus03
        assert minus03 == minus015
        assert minus015 != zero
        assert zero == plus015
        assert plus015 != plus03
        assert plus03 == plus039
        assert plus039 != plus047
        assert plus047 == plus059
        assert plus059 != plus069
        assert plus069 == plus1


def test_yields_from_leung_nomoto_2020():
    minus2 = sn.yields_from_leung_nomoto_2020(-2)
    minus165 = sn.yields_from_leung_nomoto_2020(-1.65)
    minus1 = sn.yields_from_leung_nomoto_2020(-1)
    minus065 = sn.yields_from_leung_nomoto_2020(-0.65)
    minus03 = sn.yields_from_leung_nomoto_2020(-0.3)
    minus015 = sn.yields_from_leung_nomoto_2020(-0.15)
    zero = sn.yields_from_leung_nomoto_2020(0.0)
    plus015 = sn.yields_from_leung_nomoto_2020(0.15)
    plus03 = sn.yields_from_leung_nomoto_2020(0.3)
    plus039 = sn.yields_from_leung_nomoto_2020(0.39)
    plus047 = sn.yields_from_leung_nomoto_2020(0.47)
    plus059 = sn.yields_from_leung_nomoto_2020(0.59)
    plus069 = sn.yields_from_leung_nomoto_2020(0.69)
    plus1 = sn.yields_from_leung_nomoto_2020(1)

    assert minus2 == minus165
    assert minus165 != minus1
    assert minus1 == minus065
    assert minus065 != minus03
    assert minus03 == minus015
    assert minus015 != zero
    assert zero == plus015
    assert plus015 != plus03
    assert plus03 == plus039
    assert plus039 != plus047
    assert plus047 == plus059
    assert plus059 != plus069
    assert plus069 == plus1


def test_yields_from_bravo_2019():
    bravo_sources = [
        sn.yields_from_bravo_2019_table3,
        sn.yields_from_bravo_2019_table4
    ]

    for bravo_dataset in bravo_sources:
        minus2 = bravo_dataset(-2)
        minus15 = bravo_dataset(-1.5)
        minus1 = bravo_dataset(-1)
        minus07 = bravo_dataset(-0.7)
        minus05 = bravo_dataset(-0.5)
        minus02 = bravo_dataset(-0.2)
        zero = bravo_dataset(0.0)
        plus025 = bravo_dataset(0.25)
        plus1 = bravo_dataset(1)

        assert minus2 == minus15
        assert minus15 != minus1
        assert minus1 == minus07
        assert minus07 != minus05
        assert minus05 == minus02
        assert minus02 != zero
        assert zero == plus025
        assert plus025 != plus1


def test_yields_from_gronow_2021():
    gronow_sources = [
        sn.yields_from_gronow_2021_table3_he,
        sn.yields_from_gronow_2021_table3_core
    ]

    for gronow_dataset in gronow_sources:
        minus2 = gronow_dataset(-2)
        minus15 = gronow_dataset(-1.5)
        minus1 = gronow_dataset(-1)
        minus05 = gronow_dataset(-0.5)
        zero = gronow_dataset(0.0)
        plus0239 = gronow_dataset(0.239)
        plus1 = gronow_dataset(1)

        assert minus2 == minus15
        assert minus15 != minus1
        assert minus1 == minus05
        assert minus05 != zero
        assert zero == plus0239
        assert plus0239 != plus1



import pytest
from pytest_mock import mocker

import numpy
import starmatrix.model
from starmatrix.model import Model
import starmatrix.settings as settings
import starmatrix.imfs as imfs
import starmatrix.functions as functions
import starmatrix.abundances as abundances
import starmatrix.dtds as dtds


def test_model_initialization():
    params = {
        "z": 0.033,
        "sol_ab": "ag89",
        "imf": "chabrier",
        "m_min": 1.1,
        "m_max": 42.0,
        "total_time_steps": 250,
        "dtd_sn": "greggio"
    }
    validated_params = settings.validate(params)
    model = Model(validated_params)

    assert isinstance(model.initial_mass_function, imfs.Chabrier)
    assert isinstance(model.context["abundances"], abundances.AndersGrevesse1989)
    assert model.context["abundances"].z == params["z"]
    assert model.mass_intervals == []
    assert model.energies == []
    assert model.sn_Ia_rates == []
    assert model.z == params["z"]
    assert model.dtd == dtds.dtd_greggio
    assert model.m_min == params["m_min"]
    assert model.m_max == params["m_max"]
    assert model.total_time_steps == params["total_time_steps"]


@pytest.fixture
def deactivate_open_files(mocker):
    """
    Fixture to disable opening of files from model
    """
    mocked_file = mocker.mock_open()
    mocker.patch.object(starmatrix.model, 'open', mocked_file)
    return mocked_file


def test_model_run(mocker):
    mocker.patch.object(Model, "explosive_nucleosynthesis")
    mocker.patch.object(Model, "create_q_matrices")
    Model(settings.default).run()

    Model.explosive_nucleosynthesis.assert_called()
    Model.create_q_matrices.assert_called()


def test_explosive_nucleosynthesis_with_logt_step(mocker, deactivate_open_files):
    mocker.spy(Model, "explosive_nucleosynthesis_step_t")
    mocker.spy(Model, "explosive_nucleosynthesis_step_logt")
    mocker.spy(Model, "explosive_nucleosynthesis_two_steps_t")
    mocker.spy(Model, "explosive_nucleosynthesis_fixed_n_steps")

    model = Model(settings.validate({"integration_step": "logt"}))
    model.explosive_nucleosynthesis()

    Model.explosive_nucleosynthesis_step_logt.assert_called_once()
    Model.explosive_nucleosynthesis_step_t.assert_not_called()
    Model.explosive_nucleosynthesis_two_steps_t.assert_not_called()
    Model.explosive_nucleosynthesis_fixed_n_steps.assert_not_called()


def test_explosive_nucleosynthesis_with_t_step(mocker, deactivate_open_files):
    mocker.spy(Model, "explosive_nucleosynthesis_step_t")
    mocker.spy(Model, "explosive_nucleosynthesis_step_logt")
    mocker.spy(Model, "explosive_nucleosynthesis_two_steps_t")
    mocker.spy(Model, "explosive_nucleosynthesis_fixed_n_steps")

    model = Model(settings.validate({"integration_step": "t"}))
    model.explosive_nucleosynthesis()

    Model.explosive_nucleosynthesis_step_t.assert_called_once()
    Model.explosive_nucleosynthesis_step_logt.assert_not_called()
    Model.explosive_nucleosynthesis_two_steps_t.assert_not_called()
    Model.explosive_nucleosynthesis_fixed_n_steps.assert_not_called()


def test_explosive_nucleosynthesis_with_two_steps_t(mocker, deactivate_open_files):
    mocker.spy(Model, "explosive_nucleosynthesis_step_t")
    mocker.spy(Model, "explosive_nucleosynthesis_step_logt")
    mocker.spy(Model, "explosive_nucleosynthesis_two_steps_t")
    mocker.spy(Model, "explosive_nucleosynthesis_fixed_n_steps")

    model = Model(settings.validate({"integration_step": "two_steps_t"}))
    model.explosive_nucleosynthesis()

    Model.explosive_nucleosynthesis_two_steps_t.assert_called_once()
    Model.explosive_nucleosynthesis_step_t.assert_not_called()
    Model.explosive_nucleosynthesis_step_logt.assert_not_called()
    Model.explosive_nucleosynthesis_fixed_n_steps.assert_not_called()


def test_explosive_nucleosynthesis_with_fixed_n_steps(mocker, deactivate_open_files):
    mocker.spy(Model, "explosive_nucleosynthesis_step_t")
    mocker.spy(Model, "explosive_nucleosynthesis_step_logt")
    mocker.spy(Model, "explosive_nucleosynthesis_two_steps_t")
    mocker.spy(Model, "explosive_nucleosynthesis_fixed_n_steps")

    model = Model(settings.validate({"integration_step": "fixed_n_steps",
                                     "integration_steps_stars_smaller_than_4Msun": 100,
                                     "integration_steps_stars_bigger_than_4Msun": 200}))
    model.explosive_nucleosynthesis()

    Model.explosive_nucleosynthesis_fixed_n_steps.assert_called_once_with(model, 200, 100)
    Model.explosive_nucleosynthesis_two_steps_t.assert_not_called()
    Model.explosive_nucleosynthesis_step_t.assert_not_called()
    Model.explosive_nucleosynthesis_step_logt.assert_not_called()


def test_explosive_nucleosynthesis_with_invalid_step(mocker, deactivate_open_files):
    mocker.spy(Model, "explosive_nucleosynthesis_step_t")
    mocker.spy(Model, "explosive_nucleosynthesis_step_logt")
    mocker.spy(Model, "explosive_nucleosynthesis_two_steps_t")
    mocker.spy(Model, "explosive_nucleosynthesis_fixed_n_steps")

    model = Model({**settings.default, **{"integration_step": "t2"}})
    with pytest.raises(ValueError):
        model.explosive_nucleosynthesis()

    Model.explosive_nucleosynthesis_step_t.assert_not_called()
    Model.explosive_nucleosynthesis_step_logt.assert_not_called()
    Model.explosive_nucleosynthesis_two_steps_t.assert_not_called()
    Model.explosive_nucleosynthesis_fixed_n_steps.assert_not_called()


def test_explosive_nucleosynthesis_step_t(mocker, deactivate_open_files):
    mocked_file = deactivate_open_files
    model_settings = settings.default_settings()
    model = Model(model_settings)
    model.explosive_nucleosynthesis_step_t()

    assert len(model.mass_intervals) == model_settings["total_time_steps"]
    assert len(model.energies) == model_settings["total_time_steps"]
    assert len(model.sn_Ia_rates) == model_settings["total_time_steps"]
    mocked_file.assert_called_once_with(f"{model_settings['output_dir']}/mass_intervals", "w+")


def test_explosive_nucleosynthesis_step_logt(mocker, deactivate_open_files):
    mocked_file = deactivate_open_files
    model_settings = settings.default_settings()
    model = Model(model_settings)
    model.explosive_nucleosynthesis_step_logt()

    assert len(model.mass_intervals) == model_settings["total_time_steps"]
    assert len(model.energies) == model_settings["total_time_steps"]
    assert len(model.sn_Ia_rates) == model_settings["total_time_steps"]
    mocked_file.assert_called_once_with(f"{model_settings['output_dir']}/mass_intervals", "w+")


def test_explosive_nucleosynthesis_two_steps_t(mocker, deactivate_open_files):
    mocked_file = deactivate_open_files
    model_settings = settings.default_settings()
    model = Model(model_settings)
    model.explosive_nucleosynthesis_two_steps_t()

    assert model.total_time_steps > 0
    assert model.total_time_steps != model_settings["total_time_steps"]
    assert len(model.mass_intervals) == model.total_time_steps
    assert len(model.energies) == model.total_time_steps
    assert len(model.sn_Ia_rates) == model.total_time_steps
    mocked_file.assert_called_once_with(f"{model_settings['output_dir']}/mass_intervals", "w+")


def test_explosive_nucleosynthesis_fixed_n_steps(mocker, deactivate_open_files):
    mocked_file = deactivate_open_files
    model = Model(settings.default)
    model.explosive_nucleosynthesis_fixed_n_steps(150, 91)

    assert model.total_time_steps == 150+91
    assert len(model.mass_intervals) == model.total_time_steps
    assert len(model.energies) == model.total_time_steps
    assert len(model.sn_Ia_rates) == model.total_time_steps
    mocked_file.assert_called_once_with(f"{settings.default['output_dir']}/mass_intervals", "w+")


def test_create_q_matrices(mocker, deactivate_open_files):
    mocker.spy(functions, "newton_cotes")
    mocked_file = deactivate_open_files
    mocker.spy(numpy, "savetxt")
    model = Model(settings.default)
    model.total_time_steps = 2
    model.mass_intervals = [[1., 8.], [8., 33.]]
    model.sn_Ia_rates = [2e-4, 1e-4]
    model.energies = [3e-4, 1.2e-4]

    model.create_q_matrices()

    calls = [mocker.call(f"{settings.default['output_dir']}/imf_supernova_rates", "w+"),
             mocker.call(f"{settings.default['output_dir']}/qm-matrices", "w+")]
    mocked_file.assert_has_calls(calls)
    functions.newton_cotes.assert_has_calls
    assert numpy.savetxt.call_count == model.total_time_steps


def test_create_q_matrices_all_zeros_files_if_wrong_data(mocker, deactivate_open_files):
    mocker.spy(functions, "newton_cotes")
    mocked_file = deactivate_open_files
    mocker.spy(numpy, "savetxt")
    model = Model(settings.default)
    model.total_time_steps = 2
    model.mass_intervals = [[0., 0.1], [8., 2.]]
    model.energies = [1e-8, -1e-4]

    model.create_q_matrices()

    calls = [mocker.call(f"{settings.default['output_dir']}/imf_supernova_rates", "w+"),
             mocker.call(f"{settings.default['output_dir']}/qm-matrices", "w+")]
    mocked_file.assert_has_calls(calls)
    assert functions.newton_cotes.call_count == 0
    assert numpy.savetxt.call_count == model.total_time_steps


def test_matrix_header():
    model = Model(settings.default)
    assert model.context["matrix_headers"] is True
    assert model._matrix_header(100, 1) == "Q matrix for mass interval: [100, 1]"

    model.context["matrix_headers"] = False
    assert model._matrix_header(100, 1) == ""


def test_return_fractions(mocker, deactivate_open_files):
    mocker.spy(functions, "return_fraction")
    mocked_file = deactivate_open_files
    model = Model({**settings.default, **{"return_fractions": True}})
    model.total_time_steps = 2
    model.mass_intervals = [[1., 8.], [8., 33.]]
    model.sn_Ia_rates = [2e-4, 1e-4]
    model.energies = [3e-4, 1.2e-4]

    model.create_q_matrices()

    assert model.context["return_fractions"] is True
    calls = [mocker.call(f"{settings.default['output_dir']}/imf_supernova_rates", "w+"),
             mocker.call(f"{settings.default['output_dir']}/qm-matrices", "w+"),
             mocker.call(f"{settings.default['output_dir']}/return_fractions", "w+")]
    mocked_file.assert_has_calls(calls)
    functions.return_fraction.assert_has_calls



#!/usr/bin/env python
from setuptools import setup

setup()

# https://packaging.python.org/guides/single-sourcing-package-version/
# https://packaging.python.org/guides/distributing-packages-using-setuptools
# http://blog.ionelmc.ro/2014/05/25/python-packaging/



import pytest
import numpy as np
import starmatrix.settings as settings
from starmatrix.abundances import Abundances, select_abundances
from starmatrix.abundances import AndersGrevesse1989, GrevesseSauval1998, Asplund2005, Asplund2009, Heger2010, Lodders2019


@pytest.fixture
def available_abundances():
    """
    Fixture returning all the valid values for the solar abundance setting
    """
    return settings.valid_values["sol_ab"]


def test_select_abundances():
    strings = ["ag89", "gs98", "as05", "as09", "he10", "lo19"]
    classes = [AndersGrevesse1989, GrevesseSauval1998, Asplund2005, Asplund2009, Heger2010, Lodders2019]

    for i in range(len(strings)):
        abundance_instance = select_abundances(strings[i], 0.033)
        assert type(abundance_instance) == classes[i]
        assert abundance_instance.z == 0.033


def test_abundances_classes_presence(available_abundances):
    for abundance in available_abundances:
        assert select_abundances(abundance, 0.02) is not None


def test_description_presence(available_abundances):
    for abundance in available_abundances:
        assert select_abundances(abundance, 0.02).description() != Abundances(0.02).description()


def test_feh_with_z_zero():
    assert Abundances(0).feh() == -20


def test_feh_with_z_non_zero(available_abundances):
    for abundance in available_abundances:
        assert select_abundances(abundance, 0.025).feh() > 0.0


def test_h_he4_values(available_abundances):
    for abundance in available_abundances:
        assert select_abundances(abundance, 0.025).h() > 0.0
        assert select_abundances(abundance, 0.025).he4() > 0.0


def test_elements_value_presence(available_abundances):
    for abundance in available_abundances:
        elements = np.array(list(select_abundances(abundance, 0.025).elements().values()))
        assert np.all(elements > 0.0)
        assert np.all(elements < 1.0)


def test_abundance_values_presence(available_abundances):
    for abundance in available_abundances:
        abundance_values = np.array(list(select_abundances(abundance, 0.025).abundance().values()))
        assert np.all(abundance_values > 0.0)


def test_cri_lim_exception(available_abundances):
    for abundance in available_abundances:
        normal = select_abundances(abundance, 0.02).abundance()
        cri_lim_corrected = select_abundances(abundance, 0.02).corrected_abundance_CRI_LIM()
        assert normal == cri_lim_corrected

        normal = select_abundances(abundance, 0.013).abundance()
        cri_lim_corrected = select_abundances(abundance, 0.013).corrected_abundance_CRI_LIM()
        assert normal != cri_lim_corrected
